from dinora.utils import disable_tensorflow_log

disable_tensorflow_log()
from tensorflow import keras

from dinora.net import ChessModel
from dinora.dataset import create_dataset_from_games
from dinora.model import build_model, LightConfig
from dinora.selfplay import gen_game

nodes = 5
net = ChessModel("models/best_light_model.h5")
net.model = build_model(LightConfig)  # Start from zero
net.model.compile(
    keras.optimizers.Adam(0.05),
    loss=["categorical_crossentropy", "mean_squared_error"],
    metrics=["accuracy"],
)

c_puct = 2.0
softmax_temp = 1.6
dirichlet_alpha = 0.3
noise_eps = 0.2

stats = {}
i = 0
while True:
    selfplay_games = [
        gen_game(nodes, net, c_puct, dirichlet_alpha, noise_eps, softmax_temp)
        for _ in range(3)
    ]

    dataset = create_dataset_from_games(selfplay_games)
    dataset = dataset.batch(4096)
    net.model.fit(dataset, epochs=1)

    for g in selfplay_games:
        res = g.headers["Result"]
        stats[res] = stats.setdefault(res, 0) + 1

    if i % 256 == 0:
        net.model.save_weights("weights.h5")
        print(stats)
        print(selfplay_games[0])
    i += 1



from itertools import product
from statistics import mean

import chess
import chess.engine
from chess.pgn import Game

from .utils import disable_tensorflow_log

disable_tensorflow_log()
from .search import uct_nodes
from .net import ChessModel

output_filename = "games.pgn"
number_of_games = 2
elo_k = 32

stockfish_elo = [1350]  # [1400, 1600, 1800]
stockfish_nodes = 100_000

dinora_nodes = 50
dinora_c = 2.0
dinora_net = ChessModel("models/best_light_model.h5")
dinora_elo = mean(stockfish_elo)
dinora_name = "Dinora"

fish = chess.engine.SimpleEngine.popen_uci("stockfish")
fish.configure({"UCI_LimitStrength": True, "UCI_Elo": 1400})
output_file = open(output_filename, "w", encoding="ascii")

color = True
for fish_elo, game_n in product(stockfish_elo, range(number_of_games)):
    stockfish_name = f"<Stockfish {fish_elo} nodes {stockfish_nodes}>"
    fish.configure(
        {
            "UCI_Elo": fish_elo,
        }
    )
    dinora_color = color
    if dinora_color:
        white = dinora_name
        black = stockfish_name
    else:
        white = stockfish_name
        black = dinora_name

    print(f"Dinora Elo: {dinora_elo}")
    print(f"{white} vs {black}")

    game = Game()
    game.headers["Event"] = "Dinora Elo Estimating"
    game.headers["White"] = white
    game.headers["Black"] = black

    node = game
    while not node.board().is_game_over(claim_draw=True):
        if dinora_color:
            # move = fish.play(node.board(), limit=chess.engine.Limit(nodes=100_000)).move
            move, _ = uct_nodes(node.board(), dinora_nodes, dinora_net, dinora_c)
            move = chess.Move.from_uci(move)
        else:
            move = fish.play(node.board(), limit=chess.engine.Limit(nodes=100_000)).move
        node = node.add_variation(move)
        dinora_color = not dinora_color
        print(".", end="", flush=True)
    print()  # newline after dots

    result = node.board().result(claim_draw=True)
    game.headers["Result"] = result
    print(game, file=output_file, end="\n\n")

    # Update Dinora Elo rating
    dinora_estimated = 1 / (1 + 10 ** ((fish_elo - dinora_elo) / 400))
    if result == "*":
        raise ValueError("Game is broken")
    elif (result == "1-0" and color) or (result == "0-1" and not color):
        score = 1.0
    elif result == "1/2-1/2":
        score = 0.5
    else:
        score = 0.0
    dinora_elo = dinora_elo + elo_k * (score - dinora_estimated)

    print(f"Dinora game score: {score}")
    print(f"Dinora new Elo: {dinora_elo}\n")
    color = not color

print(f"Final Dinora Elo: {dinora_elo}")
fish.quit()
output_file.close()



import tensorflow as tf

from .preprocess_pgn import load_chess_games, chess_positions
from .selfplay import gen_game

default_dinora_signature = (
    tf.TensorSpec(shape=(18, 8, 8), dtype=tf.float32, name=None),  # Board
    (
        tf.TensorSpec(shape=(1968,), dtype=tf.float32, name=None),  # Policy
        tf.TensorSpec(shape=(), dtype=tf.float32, name=None),  # Game result
    ),
)


def create_dataset_from_pgn(pgn_path: str, max_games: int) -> tf.data.Dataset:
    games = load_chess_games(pgn_path, max_games)
    return create_dataset_from_games(games)


def create_dataset_from_selfplay(nodes, net, c) -> tf.data.Dataset:
    games = [gen_game(nodes, net, c)]
    return create_dataset_from_games(games)


def create_dataset_from_games(games) -> tf.data.Dataset:
    tfdataset = tf.data.Dataset.from_generator(
        lambda: chess_positions(games), output_signature=default_dinora_signature
    )
    return tfdataset



import logging

import chess
import chess.pgn

from .policy import policy_from_move, flip_policy
from .board_representation.canon_planes import canon_input_planes

logging.basicConfig(level=logging.DEBUG)


def num_result(game):
    result = game.headers["Result"]
    if result == "0-1":
        result = -1.0
    elif result == "1-0":
        result = 1.0
    elif result == "1/2-1/2":
        result = 0.0
    else:
        raise ValueError(f"Illegal game result: {result}")
    return result


def planes_input(fen: str, flip: bool):
    plane = canon_input_planes(fen, flip)
    return plane


def result_output(result: float, flip: bool):
    return -result if flip else result


def policy_output(move: chess.Move, flip: bool):
    policy = policy_from_move(move)
    if flip:
        policy = flip_policy(policy)
    return policy


def load_chess_games(filename_pgn: str, max_games: int):
    pgn = open(filename_pgn, "r", encoding="utf8", errors="ignore")
    game = True
    i = 0
    while game:
        game = chess.pgn.read_game(pgn)
        if not game or game.headers.get("Variant", "Standard") != "Standard":
            game = chess.pgn.read_game(pgn)
            continue
        i += 1
        if i > max_games:
            break
        if i % 1000 == 0:
            print(i)
        yield game


def chess_positions(games):
    for game in games:
        try:
            result = num_result(game)
        except ValueError:
            continue
        board = game.board()

        for move in game.mainline_moves():
            fen = board.fen()
            flip = not board.turn
            yield (
                planes_input(fen, flip),
                (policy_output(move, flip), result_output(result, flip)),
            )

            try:
                board.push(move)
            except AssertionError:
                logging.warning(
                    "Broken game found," f"can't make a move {move}." "Skipping"
                )



import chess
from chess.pgn import Game

from .utils import disable_tensorflow_log

disable_tensorflow_log()
from .search import uct_nodes


def gen_game(nodes: int, net, c_puct, dirichlet_alpha, noise_eps, softmax_temp):
    game = Game()
    node = game
    while not node.board().is_game_over(claim_draw=True):
        move, _ = uct_nodes(
            node.board(),
            nodes,
            net,
            c_puct,
            None,
            dirichlet_alpha,
            noise_eps,
            softmax_temp,
        )
        move = chess.Move.from_uci(move)
        node = node.add_variation(move)
        print(move, end=" ", flush=True)
    print(node.board().result(claim_draw=True))

    result = node.board().result(claim_draw=True)
    game.headers["Result"] = result
    return game


if __name__ == "__main__":
    from .net import ChessModel

    nodes = 10
    net = ChessModel("models/best_light_model.h5")
    c = 2.0
    game = gen_game(nodes, net, c)
    print(game)



"""
The main function here is canon_input_planes.
It takes board fen, like this: rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1
and returns a numpy array of shape (18, 8, 8) for keras CNN input in NCHW format.

N is the number of boards (since this func process only one board, it is not in numpy shape for now)
C [channels] is the number of planes (12 for pieces, 4 for castling, 1 for fifty move, 1 for en passant)
H [height] and W [width] are the chess board size 8 x 8

More about channels layers:
    Pieces layers (12), each layer is a 8 x 8 matrix of floats, with 1.0 for a piece and 0.0 for empty cell:
        0 => white king
        1 => white queen
        2 => white rook
        3 => white bishop
        4 => white knight
        5 => white pawn
        6 => black king
        7 => black queen
        8 => black rook
        9 => black bishop
        10 => black knight
        11 => black pawn
    Castling layers (4), entire layer is filled with 1.0 if castling is available, 0.0 otherwise:
        12 => white can castle kingside
        13 => white can castle queenside
        14 => black can castle kingside
        15 => black can castle queenside
    Fifty move layer (1), filled with the number of half moves since the last pawn move or capture:
        16 => fifty move count
    En passant layer (1):
        17 => en passant square

Array weight is 18 * 8 * 8 * 4 = 4608 bytes, where 4 is float32
"""

import numpy as np

pieces_order = "KQRBNPkqrbnp"  # 12x8x8
ind = {pieces_order[i]: i for i in range(12)}


def canon_input_planes(fen, flip: bool):
    """
    :param fen:
    :return : (18, 8, 8) representation of the game state
    """
    if flip:
        fen = flip_fen(fen)
    return all_input_planes(fen)


def flip_fen(fen):
    foo = fen.split(" ")
    rows = foo[0].split("/")

    def swapcase(a):
        if a.isalpha():
            return a.lower() if a.isupper() else a.upper()
        return a

    def swapall(aa):
        return "".join([swapcase(a) for a in aa])

    return (
        "/".join([swapall(row) for row in reversed(rows)])
        + " "
        + ("w" if foo[1] == "b" else "b")
        + " "
        + "".join(sorted(swapall(foo[2])))
        + " "
        + foo[3]
        + " "
        + foo[4]
        + " "
        + foo[5]
    )


def all_input_planes(fen):
    history_both = to_planes(fen)  # [12*8*8]
    current_aux_planes = aux_planes(fen)  # [6*8*8]

    ret = np.vstack((history_both, current_aux_planes))
    assert ret.shape == (18, 8, 8)
    return ret


def aux_planes(fen):
    foo = fen.split(" ")

    en_passant = np.zeros((8, 8), dtype=np.float32)
    if foo[3] != "-":
        eps = alg_to_coord(foo[3])
        en_passant[eps[0]][eps[1]] = 1

    fifty_move_count = int(foo[4])
    fifty_move = np.full((8, 8), fifty_move_count, dtype=np.float32)

    castling = foo[2]
    auxiliary_planes = [
        np.full((8, 8), int("K" in castling), dtype=np.float32),
        np.full((8, 8), int("Q" in castling), dtype=np.float32),
        np.full((8, 8), int("k" in castling), dtype=np.float32),
        np.full((8, 8), int("q" in castling), dtype=np.float32),
        fifty_move,
        en_passant,
    ]

    ret = np.asarray(auxiliary_planes, dtype=np.float32)
    assert ret.shape == (6, 8, 8)
    return ret


def alg_to_coord(alg):
    rank = 8 - int(alg[1])  # 0-7
    file = ord(alg[0]) - ord("a")  # 0-7
    return rank, file


def to_planes(fen):
    board_state = replace_tags_board(fen)
    pieces_both = np.zeros(shape=(12, 8, 8), dtype=np.float32)
    for rank in range(8):
        for file in range(8):
            v = board_state[rank * 8 + file]
            if v.isalpha():
                pieces_both[ind[v]][rank][file] = 1
    assert pieces_both.shape == (12, 8, 8)
    return pieces_both


def replace_tags_board(board_san):
    board_san = board_san.split(" ")[0]
    board_san = board_san.replace("2", "11")
    board_san = board_san.replace("3", "111")
    board_san = board_san.replace("4", "1111")
    board_san = board_san.replace("5", "11111")
    board_san = board_san.replace("6", "111111")
    board_san = board_san.replace("7", "1111111")
    board_san = board_san.replace("8", "11111111")
    return board_san.replace("/", "")



from tensorflow import keras
from tensorflow.keras import layers  # type: ignore
from tensorflow.keras.regularizers import L2  # type: ignore

# Ignore Pylance reportMissingImports


class ModelConfig:
    cnn_filter_num = 256
    cnn_first_filter_size = 5
    cnn_filter_size = 3
    res_layer_num = 7
    l2_reg = 1e-4


class LightConfig(ModelConfig):
    cnn_filter_num = 128
    cnn_first_filter_size = 3
    cnn_filter_size = 3
    res_layer_num = 10
    l2_reg = 1e-4


def build_model(mc: ModelConfig) -> keras.Model:
    base_inputs, base_outputs = build_base_block(mc)
    res_outputs = build_residual_blocks(base_outputs, mc)
    policy_out = build_policy_output(res_outputs, mc)
    value_out = build_value_output(res_outputs, mc)

    model = keras.Model(base_inputs, [policy_out, value_out], name="chess_model")
    model.compile(
        optimizer="adam",
        loss=["categorical_crossentropy", "mean_squared_error"],
        metrics=["accuracy"],
    )
    return model


def build_base_block(mc: ModelConfig):
    # input_shape = (batch, channels, height, columns)
    inputs = layers.Input(shape=(18, 8, 8))
    x = layers.Conv2D(
        filters=mc.cnn_filter_num,
        kernel_size=mc.cnn_first_filter_size,
        padding="same",
        data_format="channels_first",
        use_bias=False,
        kernel_regularizer=L2(mc.l2_reg),
        name=f"input_Conv2d-{mc.cnn_first_filter_size}-{mc.cnn_filter_num}",
    )(inputs)
    x = layers.BatchNormalization(axis=1, name="input_batchnorm")(x)
    x = layers.Activation("relu", name="input_relu")(x)
    return inputs, x


def build_residual_blocks(inputs, mc: ModelConfig):
    x = inputs
    for i in range(mc.res_layer_num):
        x = build_residual_block(x, i + 1, mc)
    return x


def build_residual_block(inputs, index, mc: ModelConfig):
    res_name = f"res-{index}"
    x = layers.Conv2D(
        filters=mc.cnn_filter_num,
        kernel_size=mc.cnn_filter_size,
        padding="same",
        data_format="channels_first",
        use_bias=False,
        kernel_regularizer=L2(mc.l2_reg),
        name=f"{res_name}_conv2-{mc.cnn_filter_size}-{mc.cnn_filter_num}",
    )(inputs)
    x = layers.BatchNormalization(axis=1, name=f"{res_name}_batchnorm1")(x)
    x = layers.Activation("relu", name=f"{res_name}_relu1")(x)
    x = layers.Conv2D(
        filters=mc.cnn_filter_num,
        kernel_size=mc.cnn_filter_size,
        padding="same",
        data_format="channels_first",
        use_bias=False,
        kernel_regularizer=L2(mc.l2_reg),
        name=f"{res_name}_conv2-256-3",
    )(x)
    x = layers.BatchNormalization(axis=1, name=f"res{index}_batchnorm2")(x)
    x = layers.add([x, inputs])
    x = layers.Activation("relu", name=f"{res_name}_relu2")(x)
    return x


def build_policy_output(inputs, mc: ModelConfig):
    x = layers.Conv2D(
        filters=2,
        kernel_size=1,
        data_format="channels_first",
        use_bias=False,
        kernel_regularizer=L2(mc.l2_reg),
        name="policy_conv-1-2",
    )(inputs)
    x = layers.BatchNormalization(axis=1, name="policy_batchnorm")(x)
    x = layers.Activation("relu", name="policy_relu")(x)
    x = layers.Flatten(name="policy_flatten")(x)
    x = layers.Dense(
        1968, kernel_regularizer=L2(mc.l2_reg), activation="softmax", name="policy_out"
    )(x)
    return x


def build_value_output(inputs, mc: ModelConfig):
    x = layers.Conv2D(
        filters=4,
        kernel_size=1,
        data_format="channels_first",
        use_bias=False,
        kernel_regularizer=L2(1e-4),
        name="value_conv-1-4",
    )(inputs)
    x = layers.BatchNormalization(axis=1, name="value_batchnorm")(x)
    x = layers.Activation("relu", name="value_relu")(x)
    x = layers.Flatten(name="value_flatten")(x)
    x = layers.Dense(
        256, kernel_regularizer=L2(mc.l2_reg), activation="relu", name="value_dense"
    )(x)
    x = layers.Dense(
        1, kernel_regularizer=L2(mc.l2_reg), activation="tanh", name="value_out"
    )(x)
    return x


if __name__ == "__main__":
    model = build_model(ModelConfig)
    model.summary()

    lmodel = build_model(LightConfig)
    lmodel.summary()



import math
from os.path import dirname, realpath, join

import chess
import pylru
from tensorflow import keras
import numpy as np

from .board_representation.canon_planes import canon_input_planes
from .policy import move_lookup, flipped_move_lookup


MODELS_DIR = join(dirname(realpath(__file__)), "..", "models")
BEST_MODEL = join(MODELS_DIR, "best_model.h5")


class ChessModel:
    def __init__(self, model_path=BEST_MODEL) -> None:
        self.model: keras.Model = keras.models.load_model(model_path)

    def model_out(self, board: chess.Board):
        flip = not board.turn
        plane = canon_input_planes(board.fen(), flip)
        plane = np.array([plane])
        model_output = self.model(plane, training=False)
        return model_output

    def raw_eval(self, board: chess.Board, softmax_temp):
        policy, value = self.model_out(board)
        # unwrap tensor
        policy = policy[0]
        value = float(value[0])

        # take only legal moves from policy
        t = softmax_temp
        moves = []
        policies = []
        lookup = move_lookup if board.turn else flipped_move_lookup
        for move in board.legal_moves:
            i = lookup[move]
            moves.append(move.uci())
            policies.append(math.exp(float(policy[i]) / t))

        # map to sum(policies) == 1
        s = sum(policies)
        policies = map(lambda e: math.exp(e / t) / s, policies)

        return dict(zip(moves, policies)), value

    def evaluate(self, board: chess.Board, softmax_temp):
        result = None
        if board.is_game_over(claim_draw=True):
            result = board.result(claim_draw=True)

        if result != None:
            if result == "1/2-1/2":
                return dict(), 0.0
            else:
                # Always return -1.0 when checkmated
                # and we are checkmated because it's our turn to move
                return dict(), -1.0
        return self.raw_eval(board, softmax_temp)


class ChessModelWithCache:
    def __init__(self, size=200000, model_path=BEST_MODEL):
        self.cache = pylru.lrucache(size)
        self.net = ChessModel(model_path=model_path)

    def evaluate(self, board: chess.Board, softmax_temp):
        epd = board.epd()
        if epd in self.cache:
            policy, value = self.cache[epd]
            return policy, value
        else:
            policy, value = self.net.evaluate(board, softmax_temp)
            self.cache[epd] = [policy, value]
            return policy, value


if __name__ == "__main__":

    def print_policy(p):
        items = list(p.items())
        items.sort(key=lambda x: x[1])
        for item in items:
            print(item)

    fen = input("fen>")
    board = chess.Board(fen)
    net = ChessModel()

    policy, score = net.evaluate(board)
    print_policy(policy)
    print(score)

    import badgyal

    bad = badgyal.BGNet(False)
    policy, score = bad.eval(board)
    print_policy(policy)
    print(score)



import sys
import traceback

import chess

from dinora.utils import disable_tensorflow_log

disable_tensorflow_log()
from dinora import search

from math import cos


extra_time = 0.5
c_puct = 2.0
softmax_temp = 1.6
dirichlet_alpha = 0.3
noise_eps = 0.00


def send(s: str):
    sys.stdout.write(s)
    sys.stdout.write("\n")
    sys.stdout.flush()


class UciState:
    def __init__(self):
        self.net = None
        self.board = chess.Board()

    def load_neural_network(self):
        if self.net is None:
            send("info string loading nn, it make take a while")
            import dinora.net

            self.net = dinora.net.ChessModel("models/latest.h5")
            send("info string nn is loaded")


def time_manager(moves_number: int, time_left: int, inc: int = 0) -> float:
    moves_left = (23 * cos(moves_number / 25) + 26) / (0.01 * moves_number + 1)
    remaining_time = time_left / 1000 + moves_left * inc / 1000
    move_time = remaining_time / moves_left - extra_time
    return move_time


def uci_command(state: UciState, cmd: str):
    tokens = cmd.split()
    if tokens[0] == "uci":
        send("id name Dinora Docker")
        send("id author Saegl")
        send("option name model type string default models/latest.h5")
        send("uciok")
    elif tokens[0] == "isready":
        state.load_neural_network()
        send("readyok")
    elif tokens[0] == "ucinewgame":
        pass
    elif tokens[0] == "position":
        if tokens[1] == "startpos":
            state.board = chess.Board()
        if tokens[1] == "fen":
            fen = " ".join(tokens[2:8])
            state.board = chess.Board(fen)

        if "moves" in tokens:
            index = tokens.index("moves")
            moves = tokens[index + 1 :]
            for move in moves:
                state.board.push_uci(move)
    elif tokens[0] == "go":
        state.load_neural_network()
        # go wtime <wtime> btime <btime>
        if len(tokens) == 5 and tokens[1] == "wtime":
            wtime = int(tokens[2])
            btime = int(tokens[4])
            engine_time = wtime if state.board.turn else btime
            move_time = time_manager(state.board.fullmove_number, engine_time)
            move, _ = search.uct_time(
                state.board,
                state.net,
                c_puct,
                move_time,
                send,
                dirichlet_alpha,
                noise_eps,
                softmax_temp,
            )
        # go wtime <wtime> btime <btime> winc <winc> binc <binc>
        elif len(tokens) == 9 and tokens[1] == "wtime":
            wtime = int(tokens[2])
            btime = int(tokens[4])
            winc = int(tokens[6])
            binc = int(tokens[8])
            engine_time = wtime if state.board.turn else btime
            engine_inc = winc if state.board.turn else binc
            move_time = time_manager(
                state.board.fullmove_number, engine_time, engine_inc
            )
            move, _ = search.uct_time(
                state.board,
                state.net,
                c_puct,
                move_time,
                send,
                dirichlet_alpha,
                noise_eps,
                softmax_temp,
            )
        else:
            move, _ = search.uct_nodes(
                state.board,
                300,
                state.net,
                c_puct,
                send,
                dirichlet_alpha,
                noise_eps,
                softmax_temp,
            )
        send(f"bestmove {move}")
    elif cmd == "quit":
        exit()
    else:
        send(f"info string command is not processed: {cmd}")


def start_uci():
    uci_state = UciState()

    while True:
        line = input()
        uci_command(uci_state, line)


if __name__ == "__main__":
    try:
        start_uci()
    except:
        logfile = open("dinora.log", "w")
        exc_type, exc_value, exc_tb = sys.exc_info()
        logfile.write("".join(traceback.format_exception(exc_type, exc_value, exc_tb)))
        logfile.write("\n")
        logfile.flush()
        logfile.close()



from typing import Tuple
import numpy as np
import math
import chess
from collections import OrderedDict
from time import time

FPU = -1.0
FPU_ROOT = 0.0


def cp(Q):
    return int(295 * Q / (1 - 0.976953125 * Q**14))


class UCTNode:
    def __init__(self, board=None, parent=None, move=None, prior=0):
        self.board = board
        self.move = move
        self.is_expanded = False
        self.parent = parent  # Optional[UCTNode]
        self.children = OrderedDict()  # Dict[move, UCTNode]
        self.prior = prior  # float
        if parent == None:
            self.total_value = FPU_ROOT  # float
        else:
            self.total_value = FPU
        self.number_visits = 0  # int

    def Q(self):  # returns float
        return self.total_value / (1 + self.number_visits)

    def U(self):  # returns float
        return (
            math.sqrt(self.parent.number_visits) * self.prior / (1 + self.number_visits)
        )

    def best_child(self, C):
        return max(self.children.values(), key=lambda node: node.Q() + C * node.U())

    def select_leaf(self, C):
        current = self
        while current.is_expanded and current.children:
            current = current.best_child(C)
        if not current.board:
            current.board = current.parent.board.copy()
            current.board.push_uci(current.move)
        return current

    def expand(self, child_priors, dir_alpha, noise_eps):
        self.is_expanded = True
        e = noise_eps

        if self.parent is None:  # Is root node
            noise = np.random.dirichlet([dir_alpha] * len(child_priors))

        for i, (move, prior) in enumerate(child_priors.items()):
            if self.parent is None:
                prior = (1 - e) * prior + e * noise[i]
            self.add_child(move, prior)

    def add_child(self, move, prior):
        self.children[move] = UCTNode(parent=self, move=move, prior=prior)

    def backup(self, value_estimate: float):
        current = self
        # Child nodes are multiplied by -1 because we want max(-opponent eval)
        turnfactor = -1
        while current.parent is not None:
            current.number_visits += 1
            current.total_value += value_estimate * turnfactor
            current = current.parent
            turnfactor *= -1
        current.number_visits += 1

    def __str__(self) -> str:
        return f"UCTNode <{self.board}, {self.children}>"


def get_best_move(root):
    bestmove, node = max(
        root.children.items(), key=lambda item: (item[1].number_visits, item[1].Q())
    )
    score = int(round(cp(node.Q()), 0))
    return bestmove, node, score


def send_info(send, bestmove, count, delta, score):
    if send != None:
        send(
            "info depth 1 seldepth 1 score cp {} nodes {} nps {} pv {}".format(
                score, count, int(round(count / delta, 0)), bestmove
            )
        )


def send_tree_info(send, root):
    if send != None:
        for nd in sorted(root.children.items(), key=lambda item: item[1].number_visits):
            send(
                "info string {} {} \t(P: {}%) \t(Q: {})".format(
                    nd[1].move,
                    nd[1].number_visits,
                    round(nd[1].prior * 100, 2),
                    round(nd[1].Q(), 5),
                )
            )


def uct_nodes(
    board, nodes, net, C, send, dirichlet_alpha, noise_eps, softmax_temp
) -> Tuple[chess.Move, float]:
    start = time()
    count = 0
    delta_last = 0

    root = UCTNode(board)
    for _ in range(nodes):
        count += 1
        leaf = root.select_leaf(C)
        child_priors, value_estimate = net.evaluate(leaf.board, softmax_temp)
        leaf.expand(child_priors, dirichlet_alpha, noise_eps)
        leaf.backup(value_estimate)
        now = time()
        delta = now - start
        if delta - delta_last > 5:  # Send info every 5 sec
            delta_last = delta
            bestmove, _, score = get_best_move(root)
            send_info(send, bestmove, count, delta, score)

    bestmove, _, score = get_best_move(root)

    send_tree_info(send, root)
    send_info(send, bestmove, count, delta, score)

    # if we have a bad score, go for a draw
    return bestmove, score


def uct_time(
    board, net, C, move_time, send, dirichlet_alpha, noise_eps, softmax_temp
) -> Tuple[chess.Move, float]:
    start = time()
    count = 0
    delta_last = 0

    root = UCTNode(board)

    for _ in range(2):  # Calculate min 2 nodes
        count += 1
        leaf = root.select_leaf(C)
        child_priors, value_estimate = net.evaluate(leaf.board, softmax_temp)
        leaf.expand(child_priors, dirichlet_alpha, noise_eps)
        leaf.backup(value_estimate)

    while True:
        count += 1
        leaf = root.select_leaf(C)
        child_priors, value_estimate = net.evaluate(leaf.board, softmax_temp)
        leaf.expand(child_priors, dirichlet_alpha, noise_eps)
        leaf.backup(value_estimate)

        now = time()
        delta = now - start
        if delta - delta_last > 5:
            delta_last = delta
            bestmove, _, score = get_best_move(root)
            send_info(send, bestmove, count, delta, score)

        if delta > move_time:
            break

    bestmove, _, score = get_best_move(root)

    send_tree_info(send, root)
    send_info(send, bestmove, count, delta, score)

    # if we have a bad score, go for a draw
    return bestmove, score


if __name__ == "__main__":
    from dinora.utils import disable_tensorflow_log

    disable_tensorflow_log()

    from dinora.net import ChessModel, ChessModelWithCache

    fen = input("fen> ") or chess.STARTING_FEN
    nodes = int(input("nodes> ") or 200)
    c = float(input("c> ") or 2.0)
    dirichlet_alpha = 0.3
    noise_eps = 0.25
    softmax_temp = 1.6

    board = chess.Board(fen)
    net = ChessModel("models/best_light_model.h5")

    start = time()
    bestmove, value = uct_nodes(
        board, nodes, net, c, print, dirichlet_alpha, noise_eps, softmax_temp
    )
    print("Elapsed time: ", time() - start)
    print(f"Bestmove: {bestmove}")



import click


class UnimplementedCommand(Exception):
    pass


@click.group()
def cli():
    """Dinora is a chess engine"""


@cli.command()
def uci():
    """Start UCI"""
    from dinora.uci import start_uci

    start_uci()


@cli.command()
def healthcheck():
    """Check if Dinora is installed properly"""
    raise UnimplementedCommand("Healthcheck is not yet implemented")
    # TODO check tf gpu is available
    # TODO check that keras model is reachable
    # click.secho("OK: Everything is working", fg="green")


@cli.command()
def download():
    """Download model weights"""
    raise UnimplementedCommand("Download is not yet implemented")


@cli.group()
def dev():
    """Commands for developers"""


@dev.command()
def docs():
    """This is docs"""
    print("Read the docs")


if __name__ == "__main__":
    cli()



import datetime
from tensorflow import keras


class TimeCallback(keras.callbacks.Callback):
    """Stop model training after <hours>"""

    def __init__(self, hours):
        self.endtime = datetime.datetime.now() + datetime.timedelta(hours=hours)

    def on_train_batch_begin(self, batch, logs=None):
        if datetime.datetime.now() > self.endtime:
            self.model.stop_training = True
            self.model.save("keras_model.h5")
            print("Model stopped early by TimeCallback")
            import os  # FIXME after model training is stopped, Python interpreter still running

            os._exit(
                0
            )  # Something wrong with dataset.prefetch, without it, the interpreter stops correctly



import chess
import numpy as np


def create_flipped_uci_labels():
    """
    Seems to somehow transform the labels used for describing the universal chess interface format, putting
    them into a returned list.
    :return:
    """

    def repl(x):
        return "".join([(str(9 - int(a)) if a.isdigit() else a) for a in x])

    return [repl(x) for x in create_uci_labels()]


def create_uci_labels():
    """
    Creates the labels for the universal chess interface into an array and returns them
    :return:
    """
    labels_array = []
    letters = ["a", "b", "c", "d", "e", "f", "g", "h"]
    numbers = ["1", "2", "3", "4", "5", "6", "7", "8"]
    promoted_to = ["q", "r", "b", "n"]

    for l1 in range(8):
        for n1 in range(8):
            destinations = (
                [(t, n1) for t in range(8)]
                + [(l1, t) for t in range(8)]
                + [(l1 + t, n1 + t) for t in range(-7, 8)]
                + [(l1 + t, n1 - t) for t in range(-7, 8)]
                + [
                    (l1 + a, n1 + b)
                    for (a, b) in [
                        (-2, -1),
                        (-1, -2),
                        (-2, 1),
                        (1, -2),
                        (2, -1),
                        (-1, 2),
                        (2, 1),
                        (1, 2),
                    ]
                ]
            )
            for (l2, n2) in destinations:
                if (l1, n1) != (l2, n2) and l2 in range(8) and n2 in range(8):
                    move = letters[l1] + numbers[n1] + letters[l2] + numbers[n2]
                    labels_array.append(move)
    for l1 in range(8):
        l = letters[l1]
        for p in promoted_to:
            labels_array.append(l + "2" + l + "1" + p)
            labels_array.append(l + "7" + l + "8" + p)
            if l1 > 0:
                l_l = letters[l1 - 1]
                labels_array.append(l + "2" + l_l + "1" + p)
                labels_array.append(l + "7" + l_l + "8" + p)
            if l1 < 7:
                l_r = letters[l1 + 1]
                labels_array.append(l + "2" + l_r + "1" + p)
                labels_array.append(l + "7" + l_r + "8" + p)
    return labels_array


uci_labels = create_uci_labels()
flipped_uci_labels = create_flipped_uci_labels()
unflipped_index = [uci_labels.index(x) for x in flipped_uci_labels]

move_lookup = {
    chess.Move.from_uci(move): i for move, i in zip(uci_labels, range(len(uci_labels)))
}
flipped_move_lookup = {
    chess.Move.from_uci(move): i
    for move, i in zip(flipped_uci_labels, range(len(uci_labels)))
}


def policy_from_move(move: chess.Move):
    policy = np.zeros(len(uci_labels))

    i = move_lookup[move]
    policy[i] = 1.0
    return policy


def flip_policy(pol):
    """

    :param pol policy to flip:
    :return: the policy, flipped (for switching between black and white it seems)
    """
    return np.asarray([pol[ind] for ind in unflipped_index])


if __name__ == "__main__":
    move = chess.Move.from_uci("e2e4")
    print(uci_labels)
    print(flipped_uci_labels)

    print(move_lookup[move])
    print(flipped_move_lookup[move])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dinora', 'dinora.board_representation']

package_data = \
{'': ['*']}

install_requires = \
['chess>=1.6.1,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'numpy>=1.19.2,<2.0.0',
 'pylru>=1.2.0,<2.0.0']

entry_points = \
{'console_scripts': ['dinora = dinora.cli:cli']}

setup_kwargs = {
    'name': 'dinora',
    'version': '0.0.1',
    'description': 'Dinora Chess Engine',
    'long_description': 'None',
    'author': 'Saegl',
    'author_email': 'saegl@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
